/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.definition.activity;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.pvm.Execution;
import org.jbpm.pvm.activity.ActivityExecution;
import org.jbpm.pvm.activity.ExternalActivity;
import org.jbpm.pvm.internal.env.Authentication;
import org.jbpm.pvm.internal.model.NodeImpl;
import org.jbpm.pvm.internal.model.TransitionImpl;
import org.jbpm.pvm.internal.model.VariableDefinitionImpl;
import org.jbpm.pvm.model.Condition;
import org.jbpm.pvm.model.Transition;
import org.ow2.bonita.definition.ActivityType;
import org.ow2.bonita.definition.ClassInfo;
import org.ow2.bonita.definition.JavaHook;
import org.ow2.bonita.definition.MultiInstantiator;
import org.ow2.bonita.definition.MultiInstantiatorDescriptor;
import org.ow2.bonita.definition.Performer;
import org.ow2.bonita.facade.exception.BonitaWrapperException;
import org.ow2.bonita.facade.exception.MultiInstantiatorInvocationException;
import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.runtime.ClassDataLoader;
import org.ow2.bonita.runtime.JoinRunTime;
import org.ow2.bonita.runtime.XpdlExecution;
import org.ow2.bonita.runtime.XpdlInstance;
import org.ow2.bonita.services.Recorder;
import org.ow2.bonita.services.info.ActivityInstanceCurrentInfo;
import org.ow2.bonita.services.util.ServiceEnvTool;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.EngineEnvTool;
import org.ow2.novabpm.util.Misc;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras, Pascal Verdage
 */
/**
 * Activity life cycle:
 * 1- when an instance is created, activity state is {@link ActivityState#INITIAL}
 * 2- when the execution arrives on the node, the activity becomes {@link ActivityState#READY}
 *  the activity state is recorded (before start)
 * 3- if the activity is Manual, a task is created.
 * - when the start task is started, the activity state is recorded (after start).
 * 4- the activity becomes {@link ActivityState#EXECUTING}. The business logic is executed.
 * - the activity state is recorded (before stopped)
 * 5- if the activity is Manual, wait for the task to finish. (TODO: change the activity state ?)
 * - when the task is finished, the activity state is recorded (after stopped).
 * 6- the activity becomes {@link ActivityState#COMPLETED}.
 */
public abstract class AbstractActivity implements ExternalActivity {

  public static enum SplitType {
    AND, XOR
  }

  public static enum JoinType {
    AND, XOR
  }

  /** LOG */
  private static final Logger LOG = Logger.getLogger(AbstractActivity.class.getName());
  protected long dbid;

  protected String activityId;
  protected SplitType splitType;
  protected JoinType joinType;

  protected ActivityType activityType;
  protected List<VariableDefinitionImpl> variableDefinitions;
  protected Performer performer;

  protected Set<String> deadlines;

  protected List<JavaHook> javaHooks;
  // multi instantiation
  protected ClassInfo multiInstantiationClass;
  protected String multiInstantiationVariable;

  public static final String BODY_FINISHED = "bodyFinished";
  public static final String ACT_INSTANCE_FINISHED = "instFinished";


  protected AbstractActivity() {
  }

  protected AbstractActivity(final String activityName, final JoinType joinType,
      final SplitType splitType,
      final ActivityType activityType,
      final Performer performer) {

    Misc.checkArgsNotNull(activityType);
    this.activityId = activityName;
    this.joinType = joinType;
    this.splitType = splitType;
    this.activityType = activityType;
    this.performer = performer;

    if (performer != null) {
      if (!ActivityType.task.equals(activityType)) {
        throw new BonitaRuntimeException("Activity is automatic and a performer is specified: incorrect.");
      }
    } else {
      if (ActivityType.task.equals(activityType)) {
        throw new BonitaRuntimeException("startMode or finishMode is Manual and no performer is specified: incorrect.");
      }
    }
  }

  public void execute(final ActivityExecution execution) {
    XpdlExecution xpdlExecution = (XpdlExecution) execution;
    // If instance is ended, don't execute next node.
    if (xpdlExecution.getXpdlInstance().getInstanceState().equals(InstanceState.COMPLETED)) {
      if (LOG.isLoggable(Level.FINE)) {
        LOG.fine("XPDLInstance ended : " + xpdlExecution.getXpdlInstance());
      }
      xpdlExecution.end();
      final XpdlExecution parent = xpdlExecution.getParent();
      if (parent != null) {
        parent.removeExecution(xpdlExecution);
      }
      return;
    }
    // Execute node.
    boolean joinOK = true;
    joinOK = executeJoin(execution);
    if (joinOK) {
      if (LOG.isLoggable(Level.FINE)) {
        LOG.fine("Join for activity " + this + " is OK.");
      }
      if (LOG.isLoggable(Level.FINE)) {
        LOG.fine("Crating a new iteration on activity : " + this);
      }
      createNewIteration(xpdlExecution);

      final String nodeName = execution.getNode().getName();

      if (LOG.isLoggable(Level.FINE)) {
        LOG.fine("Executing node: " + nodeName + ", class = " + this.getClass().getSimpleName());
      }

      if (this.multiInstantiationClass != null) {
        if (LOG.isLoggable(Level.FINE)) {
          LOG.fine("MultiInstantiation not null on activity " + this);
        }
        MultiInstantiatorDescriptor actInstDescr = null;

        final MultiInstantiator actInstantiator =
          ClassDataLoader.getInstance(MultiInstantiator.class,
              xpdlExecution.getXpdlInstance().getPackageDefinitionUUID(), this.multiInstantiationClass);
        try {
          actInstDescr = EngineEnvTool.getHookExecutor()
            .executeMultiInstantiator(xpdlExecution, this.activityId, actInstantiator);
          if (actInstDescr == null) {
            throw new BonitaRuntimeException("MultiInstantiator execution returned null in activity " + this.activityId);
          }
        } catch (final Exception e) {
          throw new BonitaWrapperException(
              new MultiInstantiatorInvocationException(this.multiInstantiationClass.toString(), e)
          );
        }

        xpdlExecution.setWaitingForActivityInstanceNb(actInstDescr.getJoinNumber());
        int childId = 0;
        for (final Object value : actInstDescr.getVariableValues()) {
          XpdlExecution childExec = createChildExecution(xpdlExecution, childId);
          childExec = (XpdlExecution) childExec.createScope(xpdlExecution.getNode());
          // TODO check if values are compatibles with variable type
          childExec.setVariable(this.multiInstantiationVariable, value);
          childExec.setActivityInstanceId(Integer.toString(childId));
          executeActivityInstance(childExec);
          childId++;
        }
      } else {
        xpdlExecution = (XpdlExecution) xpdlExecution.createScope(xpdlExecution.getNode());
        executeActivityInstance(xpdlExecution);
      }
    } else {
      xpdlExecution.end();
      final XpdlExecution parent = xpdlExecution.getParent();
      if (parent != null) {
        parent.removeExecution(xpdlExecution);
      }
    }
  }

  protected void executeActivityInstance(final XpdlExecution xpdlExecution) {
    xpdlExecution.setCurrentActivityInstanceUUID(
        ServiceEnvTool.getUUIDGenerator().
        getActivityInstanceUUID(xpdlExecution.getXpdlInstance().getUUID(), this.activityId));

    final Recorder recorder = ServiceEnvTool.getRecorder();
    final XpdlInstance xpdlInstance = xpdlExecution.getXpdlInstance();
    final ActivityInstanceCurrentInfo activityInstanceCurrentInfo = new ActivityInstanceCurrentInfo(
        xpdlInstance.getPackageDefinitionUUID(),
        xpdlInstance.getProcessDefinitionUUID(),
        xpdlInstance.getUUID(),
        xpdlExecution.getCurrentActivityInstanceUUID(),
        this.getActivityType(),
        this.getActivityId(),
        xpdlExecution.getIterationId(),
        xpdlExecution.getActivityInstanceId(),
        xpdlExecution.getScopeVariables());
    recorder.recordEnterActivity(activityInstanceCurrentInfo);


    final boolean canContinue = executeBody(xpdlExecution);
    if (canContinue) {
      end(xpdlExecution);
    } else {
      xpdlExecution.waitForSignal();
    }
  }

  protected void end(final XpdlExecution xpdlExecution) {
    final ActivityInstanceCurrentInfo activityInstanceCurrentInfo = new ActivityInstanceCurrentInfo(
        xpdlExecution.getXpdlInstance().getPackageDefinitionUUID(),
        xpdlExecution.getXpdlInstance().getProcessDefinitionUUID(),
        xpdlExecution.getXpdlInstance().getUUID(),
        xpdlExecution.getCurrentActivityInstanceUUID(),
        this.getActivityType(),
        this.getActivityId(),
        xpdlExecution.getIterationId(),
        xpdlExecution.getActivityInstanceId(),
        xpdlExecution.getScopeVariables());

    ServiceEnvTool.getRecorder().recordBodyEnded(activityInstanceCurrentInfo);

    if (!XpdlExecution.MAIN_INSTANCE_NAME.equals(xpdlExecution.getActivityInstanceId())) {
      // TODO: check variable values for transition conditions
      final XpdlExecution activityInstanceExecution = (XpdlExecution) xpdlExecution.destroyScope(xpdlExecution.getNode());
      activityInstanceExecution.end();
      final XpdlExecution parent = activityInstanceExecution.getParent();
      parent.removeExecution(activityInstanceExecution);
      signal(parent, ACT_INSTANCE_FINISHED, null);
    } else {
      executeSplit(xpdlExecution, true);
    }
  }

  public void signal(final ActivityExecution execution, final String signal, final Map<String, Object> parameters) {
    final XpdlExecution xpdlExecution = (XpdlExecution) execution;
    if (BODY_FINISHED.equals(signal)) {
      end(xpdlExecution);
    } else if (ACT_INSTANCE_FINISHED.equals(signal)) {
      xpdlExecution.setWaitingForActivityInstanceNb(xpdlExecution.getWaitingForActivityInstanceNb() - 1);
      if (xpdlExecution.getWaitingForActivityInstanceNb() == 0) {
        // TODO: cancel remaining activity instances
        executeSplit(xpdlExecution, false);
      }
    } else if (signal != null && this.deadlines != null) {
      for (final String deadline : this.deadlines) {
        if (deadline.equals(signal)) {
          Authentication.setUserId("SYSTEM");
          // By default, a deadline does not propagate execution
          xpdlExecution.waitForSignal();
          final String activityId = xpdlExecution.getNode().getName();
          final JavaHook javaHook = new JavaHook(signal, JavaHook.Type.onDeadline, true);
          EngineEnvTool.getHookExecutor()
            .executeHook(xpdlExecution, activityId, javaHook);
          return;
        }
      }
    }
  }

  protected void executeSplit(final Execution execution, final boolean removeScope) {
    XpdlExecution xpdlExecution = (XpdlExecution) execution;
    final NodeImpl currentNode = xpdlExecution.getNode();
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("node = " + currentNode.getName() + " - splitType = "
          + this.splitType + " - execution = " + execution.getName());
    }
    final List<Transition> transitions = currentNode.getOutgoingTransitions();
    if (transitions == null) {
      if (LOG.isLoggable(Level.FINE)) {
        LOG.fine("node = " + currentNode.getName() + " - splitType = "
            + this.splitType + " - execution = " + execution.getName() + " no transition available. Ending execution");
      }
      xpdlExecution.end();
      final XpdlExecution parent = xpdlExecution.getParent();
      if (parent != null) {
        parent.removeExecution(xpdlExecution);
      }
    } else {
      final List<Transition> transitionsToTake = new ArrayList<Transition>();
      for (final Transition t : transitions) {
        if (evaluateTransition((TransitionImpl) t, xpdlExecution)) {
          transitionsToTake.add(t);
        }
      }
      // remove not propagated variables
      if (removeScope) {
        xpdlExecution = (XpdlExecution) xpdlExecution.destroyScope(currentNode);
      }
      if (transitionsToTake.size() == 0) {
        xpdlExecution.end();
        final XpdlExecution parent = xpdlExecution.getParent();
        if (parent != null) {
          parent.removeExecution(xpdlExecution);
        }
      } else if (transitionsToTake.size() == 1  || this.splitType.equals(SplitType.XOR)) {
        // We are in a Split/AND and only one transition is true,
        // or we are in a Split/XOR, so we take the first one that is true.
        final Transition t = transitionsToTake.get(0);
        if (LOG.isLoggable(Level.FINE)) {
          LOG.fine("Taking transition " + t);
        }
        xpdlExecution.take(t);
      } else {
        // We are in a Split/AND and more than one transition is true.
        int childId = 0;
        for (final Transition t : transitionsToTake) {
          final XpdlExecution childExecution = createChildExecution(xpdlExecution, childId++);
          if (LOG.isLoggable(Level.FINE)) {
            LOG.fine("Execution " + childExecution.getName() + "is taking transition " + t);
          }
          childExecution.take(t);
        }
      }
    }
  }

  private boolean evaluateTransition(final TransitionImpl t, final XpdlExecution xpdlExecution) {
    // testing the guard condition
    final Condition condition = t.getCondition();
    final ConditionDefinition condDef = (ConditionDefinition) condition;
    boolean conditionOK = true;
    if (condDef != null) {
      conditionOK = condDef.evaluate(xpdlExecution);
      if (!conditionOK) {
        if (LOG.isLoggable(Level.FINE)) {
          LOG.fine("Unable to take transition: " + t.getName());
        }
        // TODO : cancel nodes on this branch : may be investigate for cleaning of executions ?
      }
    }
    return conditionOK;
  }

  protected XpdlExecution createChildExecution(final XpdlExecution parentExecution, final int childId) {
    final String executionName = parentExecution.getName() + "/" + childId;
    final XpdlExecution childExecution = (XpdlExecution) parentExecution.createExecution(executionName);
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Creating child execution with name " + childExecution.getName());
    }
    return childExecution;
  }

  protected boolean executeJoin(final Execution execution) {
    final XpdlExecution currentXpdlExecution = (XpdlExecution) execution;
    final JoinRunTime joinRunTime = getJoinRuntime(currentXpdlExecution);

    return joinRunTime.transitionArrived(
        currentXpdlExecution.getPreviousTransition().getSource().getName(),
        currentXpdlExecution.getIterationId());
  }

  private static JoinRunTime getJoinRunTime(final XpdlInstance instance, final NodeImpl node) {
    final JoinType joinType = ((AbstractActivity) node.getBehaviour()).getJoinType();
    JoinRunTime joinRunTime = instance.getJoinRunTime(node.getName());
    if (joinRunTime == null) {
      int joinIncomingTransitionNbRequired = 1;
      if (joinType.equals(JoinType.AND)) {
        joinIncomingTransitionNbRequired = node.getIncomingTransitions().size();
      }
      final Map<String, JoinRunTime.WaitingTransition> ins = new HashMap<String, JoinRunTime.WaitingTransition>();
      for (final Transition t : node.getIncomingTransitions()) {
        final String name = t.getSource().getName();
        ins.put(name, new JoinRunTime.WaitingTransition(name, null));
      }
      joinRunTime = new JoinRunTime(node.getName(), ins, joinIncomingTransitionNbRequired);
      instance.pushJoinRuntime(joinRunTime);
    }
    return joinRunTime;
  }

  private static JoinRunTime getJoinRuntime(final XpdlExecution xpdlExecution) {
    return getJoinRunTime(xpdlExecution.getXpdlInstance(), xpdlExecution.getNode());
  }



  public static class IterationDescriptor {
    private String joinNodeName;
    private Set<String> transitionSourceNodes;

    protected IterationDescriptor() { }
    public IterationDescriptor(final String joinNodeName, final Set<String> transitionSourceNodes) {
      this.joinNodeName = joinNodeName;
      this.transitionSourceNodes = transitionSourceNodes;
    }
    public String getJoinNodeName() {
      return this.joinNodeName;
    }
    public Set<String> getTransitionSourceNodes() {
      return this.transitionSourceNodes;
    }
    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
      return "for join " + this.joinNodeName + ", reinitialize transitions from " + this.transitionSourceNodes;
    }
  }

  private List<IterationDescriptor>  iterationDescriptors;

  public List<IterationDescriptor> getIterationDescriptors() {
    return this.iterationDescriptors;
  }

  public void setIterationDescriptors(final List<IterationDescriptor> iteration) {
    this.iterationDescriptors = iteration;
  }

  private void createNewIteration(final XpdlExecution execution) {
    if (this.iterationDescriptors != null && !this.iterationDescriptors.isEmpty()) {
      final String iterationUUID = Misc.getUniqueId("it");
      execution.setIterationId(iterationUUID);
      final XpdlInstance instance = execution.getXpdlInstance();
      for (final IterationDescriptor it : this.iterationDescriptors) {
        final JoinRunTime jr = getJoinRunTime(instance, execution.getProcessDefinition().getNode(it.getJoinNodeName()));
        if (jr.update(iterationUUID, it.getTransitionSourceNodes())) {
          throw new BonitaRuntimeException("Error in iteration: join " + it.getJoinNodeName() + " has not been reinitialized");
        }
      }
    }
  }

  private boolean executeBody(final XpdlExecution xpdlExecution) {
    if (bodyStartAutomatically()) {
      final ActivityInstanceCurrentInfo activityInstanceCurrentInfo = new ActivityInstanceCurrentInfo(
          xpdlExecution.getXpdlInstance().getPackageDefinitionUUID(),
          xpdlExecution.getXpdlInstance().getProcessDefinitionUUID(),
          xpdlExecution.getXpdlInstance().getUUID(),
          xpdlExecution.getCurrentActivityInstanceUUID(),
          this.getActivityType(),
          this.getActivityId(),
          xpdlExecution.getIterationId(),
          xpdlExecution.getActivityInstanceId(),
          xpdlExecution.getScopeVariables());

      ServiceEnvTool.getRecorder().recordBodyStarted(activityInstanceCurrentInfo);
    }
    return executeBusinessLogic(xpdlExecution);
  }

  //public abstract ActivityBody getBody(XpdlExecution xpdlExecution);

  /**
   * Return true if the execution can continue
   */
  protected abstract boolean executeBusinessLogic(Execution execution);
  protected abstract boolean bodyStartAutomatically();
  public List<VariableDefinitionImpl> getVariableDefinitions() {
    return this.variableDefinitions;
  }

  public void setVariableDefinitions(
      final List<VariableDefinitionImpl> variableDefinitions) {
    this.variableDefinitions = variableDefinitions;
  }


  public ActivityType getActivityType() {
    return this.activityType;
  }


  public Performer getPerformer() {
    return this.performer;
  }

  public JoinType getJoinType() {
    return this.joinType;
  }


  public void setJoinType(final JoinType joinType) {
    this.joinType = joinType;
  }

  public SplitType getSplitType() {
    return this.splitType;
  }

  public Set<String> getDeadlineHooks() {
    return this.deadlines;
  }

  public void setDeadlineHooks(final Set<String> deadlineHooks) {
    this.deadlines = deadlineHooks;
  }

  public void setJavaHooks(final List<JavaHook> javaHooks) {
    this.javaHooks = javaHooks;
  }


  public List<JavaHook> getJavaHooks() {
    return this.javaHooks;
  }


  public String getActivityId() {
    return this.activityId;
  }


  public void setMultiInstantiation(final String variableId, final ClassInfo instantiator) {
    this.multiInstantiationVariable = variableId;
    this.multiInstantiationClass = instantiator;
  }
}
