/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.runtime;

import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.bonita.util.BonitaRuntimeException;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class JoinRunTime {

  private static final Logger LOG = Logger.getLogger(JoinRunTime.class.getName());

  /**
   * Description of a transition which is waited by a JoinRuntime 
   * @author Guillaume Porcher
   *
   */
  public static class WaitingTransition {
    // processDefinitionUUID of the iteration for which the transition is waited
    protected String iterationUUID;
    // source node of the transition
    protected String fromNodeName;
    // is the execution already arrived by this transition
    protected boolean arrived = false;

    protected WaitingTransition() { }

    public WaitingTransition(String fromNodeName, String iterationUUID) {
      this.fromNodeName = fromNodeName;
      this.iterationUUID = iterationUUID;
    }
    public boolean isArrived() {
      return arrived;
    }
    public void setArrived(boolean arrived) {
      this.arrived = arrived;
    }
    public String getIterationUUID() {
      return iterationUUID;
    }
    public void setIterationUUID(String iterationUUID) {
      this.iterationUUID = iterationUUID;
    }
    @Override
    public String toString() {
      return fromNodeName + "(" + iterationUUID + ") -> " + arrived;
    }
  }

  protected long dbid;
  protected String nodeName;
  protected int incomingTransitionRequired;

  // map of transitions that are waited on this join
  // key is the name the transition source node.
  protected Map<String, WaitingTransition> waitingTransitions;

  protected JoinRunTime() {
  }

  public JoinRunTime(String nodeName, Map<String, WaitingTransition> waitingTransitions, int joinIncomingTransitionNbRequired) {
    this.nodeName = nodeName;
    this.waitingTransitions = waitingTransitions;
    this.incomingTransitionRequired = joinIncomingTransitionNbRequired;
  }

  /**
   * Check if the execution can continue.
   * @return true if the execution can continue (number of transitions arrived == number of transition required) 
   */
  private boolean isOk() {
    int nbArrived = 0;
    for (WaitingTransition wt : waitingTransitions.values()) {
      if (wt.isArrived()) {
        nbArrived++;
      }
    }
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Executing Join " + nodeName + ", number of required incoming calls = " + incomingTransitionRequired
          + ", still required = " + (this.incomingTransitionRequired - nbArrived));
    }
    if (LOG.isLoggable(Level.FINE)) {
      LOG.fine("Executing Join " + nodeName + ": " + waitingTransitions);
    }
    return this.incomingTransitionRequired == nbArrived;
  }

  /**
   * Signal that a new execution arrived in this join
   * @param fromName incoming transition source node name
   * @param iterationId processDefinitionUUID of the current iteration
   * @return true if the execution can continue.
   */
  public synchronized boolean transitionArrived(String fromName, String iterationId) {
    WaitingTransition fromWt = waitingTransitions.get(fromName); 
    if (fromWt != null && !fromWt.isArrived()) { 
      if (fromWt.getIterationUUID() == null || fromWt.getIterationUUID().equals(iterationId)) {
        fromWt.setArrived(true);  
      } else {
        if (LOG.isLoggable(Level.FINE)) {
          LOG.fine("transition " + fromName + "->" + nodeName 
              + " arrived with bad iterationId (" + iterationId + "/" + fromWt.getIterationUUID() + ").");
        }
        return false;
      }
    } else {
      throw new BonitaRuntimeException("transition " + fromName + "->" + nodeName + " is already arrived in this JoinRuntime.");
    }
    return isOk();
  }

  /**
   * reinitialize the join runtime (when a new iteration is started)
   * @param iterationUUID new iterationID
   * @param transitionNames source node name of waiting transitions to reinitialize
   * @return true if the joinRuntime condition is already met (join Xor with one transition arrived)
   */
  public boolean update(String iterationUUID, Set<String> transitionNames) {
    if (incomingTransitionRequired == 1) {
      // Xor entry point: reinit all incoming transitions
      // use current iterationId to block executions coming from outside
      // TODO: when Xor join cancels unwanted branches, the setIterationUUID has to be removed.
      for (WaitingTransition wt : waitingTransitions.values()) {
        wt.setArrived(false);
        wt.setIterationUUID(iterationUUID);
      }
    }
    for (String transition : transitionNames) {
      WaitingTransition wt = waitingTransitions.get(transition);
      if (wt == null) {
        throw new BonitaRuntimeException(
            "This transition is not in the list of transitions handled by this Join. "
            + "There is a problem in the cycle definition.");
      }
      wt.setArrived(false);
      wt.setIterationUUID(iterationUUID);
    }
    return isOk();
  }

  public String getNodeName() {
    return this.nodeName;
  }
  public void setNodeName(String nodeName) {
    this.nodeName = nodeName;
  }
}
