package org.ow2.bonita.runtime;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.ow2.bonita.definition.XpdlProcess;
import org.ow2.bonita.facade.runtime.InstanceState;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.services.Repository;
import org.ow2.bonita.util.BonitaRuntimeException;
import org.ow2.bonita.util.EngineEnvTool;


public class XpdlInstance {

  protected InstanceState instanceState;
  //processInstanceLevel
  protected ProcessInstanceUUID uuid;
  protected long dbid;
  protected ProcessDefinitionUUID processUUID;
  protected PackageDefinitionUUID packageUUID;
  protected XpdlExecution rootExecution;
  /**
   * List of runtime execution scopes added for runtime informations other than
   * variables (protected List<VariableScope> variableScopes;)
   */
  protected List<JoinRunTime> joinRuntimes;
  protected XpdlExecution executionToSignal;
  protected Set<XpdlInstance> childInstances;
  protected XpdlInstance parentInstance;

  protected XpdlInstance() { }

  public XpdlInstance(final XpdlProcess xpdlProcess) {
    this.processUUID = xpdlProcess.getUUID();
    this.packageUUID = xpdlProcess.getPackageDefinitionUUID();
    this.uuid = EngineEnvTool.getUUIDGenerator().getProcessInstanceUUID(this.processUUID);

    this.rootExecution = new XpdlExecution();
    this.rootExecution.xpdlInstance = this;
    // TODO: process is set in startExecution too, see how we can remove this.
    // (process is needed to assign instance name from the instance repository)
    this.rootExecution.setProcessDefinition(xpdlProcess);

    final Repository repository = EngineEnvTool.getRepository();
    repository.storeXpdlInstance(this);
    this.rootExecution.setName(this.uuid.toString());
  }

  public void setInstanceState(final InstanceState state) {
    this.instanceState = state;
  }

  public boolean isInstanceState(final InstanceState state) {
    return this.instanceState.equals(state);
  }

  public void  pushJoinRuntime(final JoinRunTime runTime) {
    if (getJoinRuntimes() == null) {
      this.joinRuntimes = new ArrayList<JoinRunTime>();
    }
    getJoinRuntimes().add(0, runTime);
  }

  public void popJoinRuntime() {
    if ((getJoinRuntimes() != null) && (!getJoinRuntimes().isEmpty())) {
      getJoinRuntimes().remove(0);
    }
  }
  public List<JoinRunTime> getJoinRuntimes() {
    return this.joinRuntimes;
  }
  public JoinRunTime getJoinRunTime(final String nodeName) {
    JoinRunTime joinRunTime = null;
    if (this.joinRuntimes != null) {
      for (final JoinRunTime jr : this.joinRuntimes) {
        if (jr.getNodeName().equals(nodeName)) {
          if (joinRunTime != null) {
            throw new BonitaRuntimeException("More than 1 joinRuntime was found for the given nodeName : " + nodeName);
          }
          joinRunTime = jr;
        }
      }
    }
    return joinRunTime;
  }

  public XpdlExecution getRootExecution() {
    return this.rootExecution;
  }

  public InstanceState getInstanceState() {
    return this.instanceState;
  }

  public ProcessInstanceUUID getUUID() {
    return this.uuid;
  }


  public long getDbid() {
    return this.dbid;
  }


  public ProcessDefinitionUUID getProcessDefinitionUUID() {
    return this.processUUID;
  }

  /**
   * @param execution
   */
  public void setExecutionToSignal(final XpdlExecution execution) {
    this.executionToSignal = execution;
    if (execution != null) {
      // executionToSignal is set to null when the subflow is ended.
      // not setting it to null can create a bug, because the executionTosignal
      // is removed from the db, but still referenced in the subflow instance.
      //
      // When executionToSignal is set to null,
      // the parent process instance has to be kept.
      // (to delete the subflow instance with the main instance)
      this.parentInstance = execution.getXpdlInstance();
      this.parentInstance.addChildInstance(this);
    }
  }

  public XpdlExecution getExecutionToSignal() {
    return this.executionToSignal;
  }

  private void addChildInstance(final XpdlInstance instance) {
    if (this.childInstances == null) {
      this.childInstances = new HashSet<XpdlInstance>();
    }
    this.childInstances.add(instance);
  }

  @Override
  public String toString() {
    String value = "Instance " + this.uuid + "(state:" + this.instanceState;
    if (this.parentInstance != null) {
      value += ", child of " + this.parentInstance;
    }
    value += ")";
    return value;
  }


  public Set<XpdlInstance> getChildInstances() {
    if (this.childInstances == null) {
      return null;
    }
    return Collections.unmodifiableSet(this.childInstances);
  }

  public XpdlInstance getParentInstance() {
    return this.parentInstance;
  }

  public PackageDefinitionUUID getPackageDefinitionUUID() {
    return this.packageUUID;
  }

}
